/// <reference types="node" />
import { Buffer } from 'buffer';
import { FieldDescription, NitroSigner, Uint, Uint64 } from '@cerc-io/nitro-util';
import { Bytes32 } from '@statechannels/nitro-protocol';
import { Signature } from '../../crypto/signatures';
import { Address } from '../../types/types';
import { Funds } from '../../types/funds';
import { FixedPart, State } from '../state/state';
import { SignedState } from '../state/signedstate';
import { Destination } from '../../types/destination';
import { Allocation } from '../state/outcome/allocation';
import { Exit, SingleAssetExit } from '../state/outcome/exit';
export declare const ErrIncorrectChannelID: Error;
export declare const ErrIncorrectTurnNum: Error;
export declare const ErrInvalidDeposit: Error;
export declare const ErrInsufficientFunds: Error;
export declare const ErrDuplicateGuarantee: Error;
export declare const ErrGuaranteeNotFound: Error;
export declare const ErrInvalidAmount: Error;
export declare const ErrNotFollower: Error;
export declare const ErrNoProposals: Error;
export declare const ErrUnsupportedQueuedProposal: Error;
export declare const ErrUnsupportedExpectedProposal: Error;
export declare const ErrNonMatchingProposals: Error;
export declare const ErrInvalidProposalSignature: Error;
export declare const ErrInvalidTurnNum: Error;
export declare const ErrNotLeader: Error;
export declare const ErrProposalQueueExhausted: Error;
export declare const ErrWrongSigner: Error;
export declare enum ProposalType {
    AddProposal = "AddProposal",
    RemoveProposal = "RemoveProposal"
}
type LedgerIndex = Uint;
export declare const Leader: LedgerIndex;
export declare const Follower: LedgerIndex;
export declare class Balance {
    destination: Destination;
    amount?: bigint;
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): Balance;
    toJSON(): any;
    constructor(params: {
        destination?: Destination;
        amount?: bigint;
    });
    asAllocation(): Allocation;
    equal(b2: Balance): boolean;
    clone(): Balance;
}
interface GuaranteeOptions {
    amount?: bigint;
    _target?: Destination;
    left?: Destination;
    right?: Destination;
}
export declare class Guarantee {
    amount?: bigint;
    _target: Destination;
    left: Destination;
    right: Destination;
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): Guarantee;
    static newGuarantee(amount: bigint | undefined, target: Destination, left: Destination, right: Destination): Guarantee;
    toJSON(): any;
    constructor(params: GuaranteeOptions);
    clone(): Guarantee;
    target(): Destination;
    equal(g2: Guarantee): boolean;
    asAllocation(): Allocation;
}
export declare class LedgerOutcome {
    private assetAddress;
    _leader: Balance;
    _follower: Balance;
    guarantees: Map<Bytes32, Guarantee>;
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): LedgerOutcome;
    toJSON(): any;
    constructor(params: {
        assetAddress?: Address;
        _leader?: Balance;
        _follower?: Balance;
        guarantees?: Map<Bytes32, Guarantee>;
    });
    leader(): Balance;
    follower(): Balance;
    static newLedgerOutcome(assetAddress: Address, leader: Balance, follower: Balance, guarantees: Guarantee[]): LedgerOutcome;
    static fromExit(sae: SingleAssetExit): LedgerOutcome;
    asOutcome(): Exit;
    _clone(): LedgerOutcome;
    clone(): LedgerOutcome;
    fundingTargets(): Destination[];
    includes(g: Guarantee): boolean;
    includesTarget(target: Destination): boolean;
}
interface VarsConstructorOptions {
    turnNum?: Uint64;
    outcome?: LedgerOutcome;
}
export declare class Vars {
    turnNum: Uint64;
    outcome: LedgerOutcome;
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): Vars;
    toJSON(): any;
    constructor(params: VarsConstructorOptions);
    asState(fp: FixedPart): State;
    clone(): Vars;
    handleProposal(p: Proposal): void;
    add(p: Add): void;
    remove(p: Remove): void;
}
interface SignedVarsConstructorOptions extends VarsConstructorOptions {
    signatures?: [Signature, Signature];
}
export declare class SignedVars extends Vars {
    signatures: [Signature, Signature];
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): SignedVars;
    toJSON(): any;
    constructor(params: SignedVarsConstructorOptions);
    clone(): SignedVars;
}
interface AddOptions extends GuaranteeOptions {
    leftDeposit?: bigint;
}
export declare class Add extends Guarantee {
    leftDeposit?: bigint;
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): Add;
    toJSON(): any;
    constructor(params: AddOptions);
    clone(): Add;
    rightDeposit(): bigint;
    equal(a2: Add): boolean;
    static newAdd(g: Guarantee, leftDeposit?: bigint): Add;
}
export declare class Remove {
    target: Destination;
    leftAmount?: bigint;
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): Remove;
    toJSON(): any;
    constructor(params: {
        target?: Destination;
        leftAmount?: bigint;
    });
    equal(r2: Remove): boolean;
    clone(): Remove;
    static newRemove(target: Destination, leftAmount?: bigint): Remove;
}
export declare class Proposal {
    ledgerID: Destination;
    toAdd: Add;
    toRemove: Remove;
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): Proposal;
    toJSON(): any;
    constructor(params: {
        ledgerID?: Destination;
        toAdd?: Add;
        toRemove?: Remove;
    });
    target(): Destination;
    clone(): Proposal;
    type(): ProposalType;
    equal(q: Proposal): boolean;
    static newAddProposal(ledgerID: Destination, g: Guarantee, leftDeposit?: bigint): Proposal;
    static newRemoveProposal(ledgerID: Destination, target: Destination, leftAmount?: bigint): Proposal;
}
type SignedProposalParams = {
    signature?: Signature;
    proposal?: Proposal;
    turnNum?: Uint64;
};
export declare class SignedProposal {
    signature: Signature;
    proposal: Proposal;
    turnNum: Uint64;
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): SignedProposal;
    toJSON(): any;
    constructor(params: SignedProposalParams);
    clone(): SignedProposal;
    channelID(): Destination;
    sortInfo(): [Destination, Uint64];
}
export declare class ConsensusChannel {
    id: Destination;
    myIndex: LedgerIndex;
    onChainFunding: Funds;
    private fp;
    private current;
    private _proposalQueue;
    static jsonEncodingMap: Record<string, FieldDescription>;
    static fromJSON(data: string): ConsensusChannel;
    toJSON(): any;
    constructor(params: {
        id?: Destination;
        myIndex?: LedgerIndex;
        onChainFunding?: Funds;
        fp?: FixedPart;
        current?: SignedVars;
        _proposalQueue?: SignedProposal[];
    });
    static newConsensusChannel(fp: FixedPart, myIndex: LedgerIndex, initialTurnNum: Uint64, outcome: LedgerOutcome, signatures: [Signature, Signature]): ConsensusChannel;
    static newLeaderChannel(fp: FixedPart, turnNum: Uint64, outcome: LedgerOutcome, signatures: [Signature, Signature]): ConsensusChannel;
    static newFollowerChannel(fp: FixedPart, turnNum: Uint64, outcome: LedgerOutcome, signatures: [Signature, Signature]): ConsensusChannel;
    fixedPart(): FixedPart;
    receive(sp: SignedProposal): void;
    isProposed(g: Guarantee): boolean;
    isProposedNext(g: Guarantee): boolean;
    consensusTurnNum(): Uint64;
    includes(g: Guarantee): boolean;
    includesTarget(target: Destination): boolean;
    hasRemovalBeenProposed(target: Destination): boolean;
    hasRemovalBeenProposedNext(target: Destination): boolean;
    isLeader(): boolean;
    isFollower(): boolean;
    leader(): Address;
    follower(): Address;
    fundingTargets(): Destination[];
    private sign;
    private recoverSigner;
    consensusVars(): Vars;
    signatures(): [Signature, Signature];
    proposalQueue(): SignedProposal[] | null;
    private latestProposedVars;
    private validateProposalID;
    participants(): Address[] | null;
    clone(): ConsensusChannel;
    supportedSignedState(): SignedState;
    unmarshalJSON(data: Buffer): void;
    private leaderReceive;
    propose(proposal: Proposal, signer: NitroSigner): Promise<SignedProposal>;
    private appendToProposalQueue;
    private followerReceive;
    signNextProposal(expectedProposal: Proposal, signer: NitroSigner): Promise<SignedProposal>;
}
export {};
